import Util from './utils/util';
import assert from 'assert';
import LightTransaction from './models/light-transaction';
import Receipt from './models/receipt';
import types from './models/types';

class Client {
  constructor (clientConfig, infinitechain) {
    this._infinitechain = infinitechain;
    this._storage = clientConfig.storage;
    this._nodeUrl = clientConfig.nodeUrl;
  }

  getProposeDeposit = (rawData = {}, address = null) => {
    return new Promise((resolve, reject) => {
      this._infinitechain.event.getProposeDeposit((err, events) => {
        if (err) {
          reject(err);
        } else {
          let metadata = rawData.metadata? rawData.metadata : null;
          let lightTxs = [];
          for (let i = 0; i < events.length; i++) {
            let logID = events[i].returnValues._dsn;
            let to = address? address : this._infinitechain.signer.getAddress();
            let value = events[i].returnValues._value;
            let assetID = events[i].returnValues._assetID;
            let lightTxData = {
              to: to,
              assetID: assetID,
              value: value,
              logID: logID,
              metadata: metadata
            };

            let lightTx = this.makeLightTx(types.deposit, lightTxData, lightTxData.metadata);
            lightTxs.push(lightTx);
          }
          resolve(lightTxs);
        }
      }, address);
    });
  }

  makeProposeDeposit = (rawData = {}, address = null) => {
    return new Promise(async (resolve, reject) => {
      this._infinitechain.event.onProposeDeposit((err, result) => {
        if (err) {
          reject(err);
        } else {
          let metadata = rawData.metadata? rawData.metadata : null;
          let logID = result.returnValues._dsn;
          let to = address? address : this._infinitechain.signer.getAddress();
          let value = result.returnValues._value;
          let assetID = result.returnValues._assetID;
          let lightTxData = {
            to: to,
            assetID: assetID,
            value: value,
            logID: logID,
            metadata: metadata
          };

          let lightTx = this.makeLightTx(types.deposit, lightTxData, lightTxData.metadata);
          resolve(lightTx);
        }
      }, address);
    });
  }

  proposeTokenDeposit = async (proposeData, privateKey = null) => {
    let contract = this._infinitechain.contract;
    let txHash = await contract.proposeTokenDeposit(proposeData, privateKey);
    return txHash;
  }
  
  getProposeWithdrawal = (rawData = {}) => {
    let address = rawData.address ? '0x' + Util.toByte32(rawData.address) : null;
    let assetID = rawData.assetID ? '0x' + Util.toByte32(rawData.assetID) : null;
    let withdrawn = rawData.withdrawn;
    return new Promise(async (resolve, reject) => {
      try {
        let booster = this._infinitechain.contract.booster();
        let wsnArrayLength = await booster.methods.getWsnArrayLength().call();
        let withdrawalLogs = [];
        for (let i = 0; i < wsnArrayLength ; i++) {
          let wsn = await booster.methods.wsnArray(i).call();
          let withdrawalLog = await booster.methods.withdrawalLogs(wsn).call();
          if (address && withdrawalLog.client != address) {
            continue;
          }
          if (assetID && withdrawalLog.assetID != assetID) {
            continue;
          }
          if (withdrawn != undefined && withdrawalLog.flag != withdrawn) {
            continue;
          }
          withdrawalLogs.push(withdrawalLog);
        }
        resolve(withdrawalLogs);
      } catch(e) {
        reject(e);
      }
    });
  }

  makeProposeWithdrawal = (rawData, address = null) => {
    assert(rawData.assetID != undefined, '\'assetID\' is not provided.');
    assert(rawData.value != undefined, '\'value\' is not provided.');
    let fee = rawData.fee? rawData.fee : '0';
    let metadata = rawData.metadata? rawData.metadata : null;
    let from = address? address : this._infinitechain.signer.getAddress();
    let lightTxData = {
      from: from,
      assetID: rawData.assetID,
      value: rawData.value,
      fee: fee,
      metadata: metadata
    };

    let lightTx = this.makeLightTx(types.withdrawal, lightTxData, lightTxData.metadata);
    return lightTx;
  }

  makeLightTx = (type, lightTxData, metadata = null) => {
    // Prepare lightTxData
    lightTxData = this._prepare(type, lightTxData);
    if (metadata) {
      if (typeof metadata == 'object') {
        metadata = JSON.stringify(metadata);
      } else {
        metadata = metadata.toString();
      }
    }
    let lightTxJson = { lightTxData: lightTxData, metadata: metadata };

    // Create lightTx
    let lightTx = new LightTransaction(lightTxJson);
    // Sign lightTx if has privateKey
    let signer = this._infinitechain.signer;
    if (signer.hasPrivateKey()) {
      lightTx = signer.signWithClientKey(lightTx);
    }
    return lightTx;
  }

  signLightTx = (lightTx, privateKey = null) => {
    if (typeof lightTx == 'string') lightTx == JSON.parse(lightTx);
    if (typeof lightTx.metadata == 'object') {
      lightTx.metadata = JSON.stringify(lightTx.metadata);
    } else {
      lightTx.metadata = lightTx.metadata.toString();
    }
    lightTx = new LightTransaction(lightTx);
    // Sign lightTx
    let signer = this._infinitechain.signer;
    let signedLightTx = signer.signWithClientKey(lightTx, privateKey);

    return signedLightTx;
  }

  sendLightTx = async (lightTx) => {
    let gringotts = this._infinitechain.gringotts;
    let verifier = this._infinitechain.verifier;
    let receipt = await gringotts.sendLightTx(lightTx);
    assert(verifier.verifyReceipt(receipt), 'Wrong signature when verify receipt received from booster.');
    return receipt;
  }

  saveReceipt = async (receipt, upload = false) => {
    assert(receipt instanceof Receipt, 'Parameter \'receipt\' should be instance of \'Receipt\'.');
    await this._storage.setReceipt(receipt.receiptHash, receipt.toJson(), upload);
  }

  getReceipt = async (receiptHash) => {
    try {
      return await this._storage.getReceipt(receiptHash);
    } catch (e) {
      throw e;
    }
  }

  getBooterBalance = async (clientAddress, assetID = null) => {
    return await this._infinitechain.gringotts.getBoosterBalance(clientAddress, assetID);
  }

  getSlice = async (stageHeight, receiptHash) => {
    return await this._infinitechain.gringotts.getSlice(stageHeight, receiptHash);
  }

  getAllReceiptHashes = async (stageHeight) => {
    return await this._storage.getReceiptHashesByStageHeight(stageHeight);
  }

  takeObjection = async (payment) => {
    return this._infinitechain.booster.takeObjection(payment);
  }

  _prepare = (type, lightTxData) => {
    assert(Object.values(types).includes(type), 'Parameter \'type\' should be one of \'deposit\', \'withdrawal\' or \'remittance\'');

    lightTxData.nonce = this._getNonce();
    switch (type) {
    case types.deposit:
      lightTxData.from = '0';
      lightTxData.fee = '0';
      break;
    case types.withdrawal:
      lightTxData.to = '0';
      lightTxData.logID = Util.sha3(Util.toByte32(lightTxData.from) + lightTxData.nonce);
      break;
    case types.remittance:
      lightTxData.logID = '0';
      break;
    }
    return lightTxData;
  }

  _getNonce () {
    return Util.sha3((Math.random()).toString());
  }

  fetchSupportedTokens = async () => {
    return await this._infinitechain.gringotts.getAssetList();
  }

  auidtReceiptSlice = async (stageHeight, receiptHash, slice) => {
    let success = false;
    let contract = this._infinitechain.contract;
    stageHeight = parseInt(stageHeight);
    // 1. Get receiptRootHash from blockchain
    let rootHashes = await contract.getStageRootHash(stageHeight);
    let receiptRootHash = rootHashes[0];
    // 2. Compute root hash
    let computedReceiptRootHash;
    if (slice.receiptHashArray.includes(receiptHash)) {
      computedReceiptRootHash = '0x' + this._computeRootHashFromSlice(slice.slice, stageHeight);
      // 3. Compare
      if (computedReceiptRootHash == receiptRootHash) {
        success = true;
      }
    }
    return success;
  }

  _computeRootHashFromSlice (slice, stageHeight) {
    let firstNode = slice.shift();

    let rootNode = slice.reduce((acc, curr) => {
      if (acc.treeNodeIndex % 2 == 0) {
        acc.treeNodeHash = Util.sha3(acc.treeNodeHash.concat(curr.treeNodeHash));
      } else {
        acc.treeNodeHash = Util.sha3(curr.treeNodeHash.concat(acc.treeNodeHash));
      }
      acc.treeNodeIndex = parseInt(acc.treeNodeIndex / 2);
      return acc;
    }, firstNode);

    return Util.sha3(rootNode.treeNodeHash + stageHeight.toString(16).padStart(64, '0'));
  }
}

export default Client;
