import assert from 'assert';
import axios from 'axios';
import LightTransaction from './models/light-transaction';
import Receipt from './models/receipt';

class Gringotts {
  constructor (config, infinitechain) {
    this._nodeUrl = config.nodeUrl;
    this._infinitechain = infinitechain;
  }

  getSlice = async (stageHeight, receiptHash) => {
    let url = this._nodeUrl + '/slice/' + stageHeight + '/' + receiptHash;
    let res = await axios.get(url);
    return res.data;
  }

  getTrees = async (stageHeight) => {
    let url = this._nodeUrl + '/trees/' + stageHeight;
    let res = await axios.get(url);
    return res.data;
  }

  sendLightTx = async (lightTx) => {
    assert(lightTx instanceof LightTransaction, 'Parameter \'lightTx\' should be instance of LightTransaction.');
    let url = this._nodeUrl + '/send/light_tx';
    let res = await axios.post(url, { lightTxJson: lightTx.toJson() });
    res = res.data;
    if (res.ok) {
      let receiptJson = res.receipt;
      let receipt = new Receipt(receiptJson);

      return receipt;
    } else {
      throw new Error(`message: ${res.message}, code: ${res.code} `);
    }
  }

  getViableStageHeight = async () => {
    assert(this._nodeUrl, 'Can not find booster node.');
    let url = this._nodeUrl + '/viable/stage/height';
    let res = await axios.get(url);
    return parseInt(res.data.height);
  }

  fetchBoosterAddress = async () => {
    let url = this._nodeUrl + '/booster/address';
    let res = await axios.get(url);
    return res.data;
  }

  fetchRootHashes = async (stageHeight = null) => {
    let url = this._nodeUrl + '/roothash';
    if (stageHeight) {
      url = url + '/' + stageHeight.toString();
    }
    let res = await axios.get(url);
    return res.data;
  }

  getOffchainReceipts = async (stageHeight, address = null) => {
    let url = this._nodeUrl + '/receipts/' + stageHeight;
    if (address) {
      address = address.padStart(40, '0');
      url = `${url}?address=${address}`;
    }
    let res = await axios.get(url);
    return res.data;
  }

  getOffchainReceiptByGSN = async (GSN) => {
    if (typeof GSN === 'number') {
      GSN = parseInt(GSN).toString(16);
    }
    let url = this._nodeUrl + '/receipt_by_gsn/' + GSN;
    let res = await axios.get(url);
    return res.data;
  }

  getAccountBalances = async (stageHeight) => {
    let url = this._nodeUrl + '/accounts/' + stageHeight;
    let res = await axios.get(url);
    return res.data;
  }

  getBoosterBalance = async (clientAddress, assetID = null) => {
    let url;
    if (!assetID) {
      url = this._nodeUrl + '/balance/' + clientAddress;
    } else {
      url = this._nodeUrl + '/balance/' + clientAddress + '?assetID=' + assetID;
    }
    let res = await axios.get(url);
    return res.data;
  }

  getAssetList = async () => {
    let url = this._nodeUrl + '/assetlist';
    let res = await axios.get(url);
    return res.data.assetList;
  }
}

export default Gringotts;
