import assert from 'assert';
import Infinitechain from './infinitechain';
import Client from './client';
import Auditor from './auditor';
import Verifier from './verifier';
import Event from './event';
import Signer from './signer';
import Gringotts from './gringotts';
import Contract from './contract';
import Memory from './storages/memory';
import Level from './storages/level';

class InfinitechainBuilder {
  setNodeUrl (url) {
    this._nodeUrl = url;
    return this;
  }

  setWeb3Url (url) {
    this._web3Url = url;
    return this;
  }

  setSignerKey (key) {
    this._signerKey = key;
    return this;
  }

  setStorage (storage, db = null) {
    if (storage == 'memory') {
      this._storage = new Memory();
    } else if (storage == 'level') {
      this._storage = new Level(db);
    } else {
      throw new Error('Not supported storage type.');
    }

    return this;
  }

  build () {
    assert(this._nodeUrl != undefined, '\'nodeUrl\' is not provided.');
    assert(this._web3Url != undefined, '\'web3Url\' is not provided.');
    assert(this._storage != undefined, '\'storage\' is not provided.');

    let clientConfig = {
      nodeUrl: this._nodeUrl,
      storage: this._storage
    };

    let auditorConfig = {
      nodeUrl: this._nodeUrl,
      storage: this._storage
    };

    let eventConfig = {
      storage: this._storage
    };

    let gringottsConfig = {
      nodeUrl: this._nodeUrl
    };

    let contractConfig = {
      web3Url: this._web3Url
    };

    let signerConfig = {};

    let verifierConfig = {};

    let infinitechain = new Infinitechain();

    let signer = new Signer(signerConfig);
    infinitechain.setSigner(signer);

    if (this._signerKey) {
      signer.importPrivateKey(this._signerKey);
    }

    let gringotts = new Gringotts(gringottsConfig, infinitechain);
    infinitechain.setGringotts(gringotts);

    let contract = new Contract(contractConfig, infinitechain);
    infinitechain.setContract(contract);

    let verifier = new Verifier(verifierConfig, infinitechain);
    infinitechain.setVerifier(verifier);

    let event = new Event(eventConfig, infinitechain);
    infinitechain.setEvent(event);

    let client = new Client(clientConfig, infinitechain);
    infinitechain.setClient(client);

    let auditor = new Auditor(auditorConfig, infinitechain);
    infinitechain.setAuditor(auditor);

    return infinitechain;
  }
}

export default InfinitechainBuilder;
