import EthUtils from 'ethereumjs-util';
import assert from 'assert';
import LightTransaction from './models/light-transaction';
import Receipt from './models/receipt';
import types from './models/types';

class Verifier {
  constructor (verifierConfig, infinitechain) {
    this.verifierConfig = verifierConfig;
    this._infinitechain = infinitechain;
  }

  verifyLightTx = (lightTx) => {
    assert(lightTx instanceof LightTransaction, 'Parameter \'lightTx\' is not a LightTransaction instance.');
    return this._verify(lightTx);
  }

  verifyReceipt = (receipt) => {
    assert(receipt instanceof Receipt, 'Parameter \'receipt\' is not a Receipt instance.');
    return this._verify(receipt);
  }

  _verify = (object) => {
    let isValid;
    let clientAddress;
    let klass;
    let isDataValid = false;
    if (object instanceof LightTransaction) {
      klass = 'lightTx';
      let target = new LightTransaction(JSON.parse(JSON.stringify(object)));
      isDataValid = target.lightTxHash === object.lightTxHash;
    } else if (object instanceof Receipt) {
      klass = 'receipt';
      let target = new Receipt(JSON.parse(JSON.stringify(object)));
      isDataValid = target.lightTxHash === object.lightTxHash && target.receiptHash === object.receiptHash;
    } else {
      throw new Error('\'object\' should be instance of \'LightTransaction\' or \'Receipt\'.');
    }

    switch (object.type()) {
    case types.deposit:
      clientAddress = object.lightTxData.to.slice(-40);
      break;
    case types.withdrawal:
    case types.remittance:
      clientAddress = object.lightTxData.from.slice(-40);
      break;
    }
    let isClientLightTxSigValid = true;
    let isBoosterReceiptSigValid = true;

    clientAddress = clientAddress.toLowerCase();

    if (object.hasClientLightTxSig()) {
      isClientLightTxSigValid = (clientAddress == this._recover(object.lightTxHash, object.sig.clientLightTx).toLowerCase());
    }

    if (klass == 'receipt') {
      let boosterAccountAddress = EthUtils.stripHexPrefix(this._infinitechain.contract.boosterAccountAddress()).toLowerCase();
      isBoosterReceiptSigValid = object.hasBoosterReceiptSig() && (boosterAccountAddress == this._recover(object.receiptHash, object.sig.boosterReceipt).toLowerCase());
    }

    isValid = (isDataValid && isClientLightTxSigValid && isBoosterReceiptSigValid);
    return isValid;
  }

  _recover = (msgHash, signature) => {
    let prefix = Buffer.from('\x19Ethereum Signed Message:\n32');
    let ethMsgHash = EthUtils.sha3(Buffer.concat([prefix, Buffer.from(msgHash, 'hex')]));
    let publicKey = EthUtils.ecrecover(ethMsgHash, signature.v, signature.r, signature.s);
    let address = EthUtils.pubToAddress(publicKey).toString('hex');
    return address;
  }
}

export default Verifier;
