import IndexedMerkleTree from '../src/utils/indexed-merkle-tree';
import Auditor from '../src/auditor';
import Util from '../src/utils/util';

import initialBalances from './test-auditor-data/initial-balances.json';
import Stage1 from './test-auditor-data/stage1-receipts.json';
import Stage2 from './test-auditor-data/stage2-receipts.json';
import receiptsWithRepeatedGSN from './test-auditor-data/repeated-GSN.json';
import receiptsWithWrongBalance from './test-auditor-data/wrong-balances.json';
import receiptsWithSkippedGSN from './test-auditor-data/skipped-GSN.json';
import receiptsWithoutIntegrity from './test-auditor-data/integrity.json';
const BN = Util.BN;

describe('Auditor', () => {
  let infinitechain;
  let stageHeight = 2;
  let address;
  let auditor;
  let receiptTree;

  beforeAll(async () => {
    receiptTree = new IndexedMerkleTree(stageHeight, Stage2.map(receipt => receipt.receiptHash));
    address = Stage2[0].lightTxData.from == 0? Stage2[0].lightTxData.to : Stage2[0].lightTxData.from;
    infinitechain = {
      contract: {
        getStageRootHash: () => { return ['0x16295c6ef33f94883a8bbe128fca32a6ad9c7006fd3cbca4d356e1fa16834b12', '0x26b4bd7104bebdfdccacc81193d9671fc9730ce624a921854dea487fa08dd36d']; }
      },
      gringotts: {
        getOffchainReceipts: (stageHeight, address = null) => {
          return Stage2.filter(receipt => {
            if (address == null) return true;
            return receipt.lightTxData.from == address || receipt.lightTxData.to == address;
          });
        },
        getAccountBalances: () => {
          return initialBalances;
        },
        getOffchainReceiptByGSN: (GSN) => {
          let allReceipts = JSON.parse(JSON.stringify(Stage1.concat(Stage2)));
          let res = {};
          for (let i = 0; i < allReceipts.length; i++) {
            if (allReceipts[i].receiptData.GSN == GSN) {
              res.data = allReceipts[i];
              break;
            }
          }
          return res;
        },
        getTrees: (stageHeight) => {
          let accountHash = Object.keys(initialBalances)
            .reduce((acc, address) => {
              let balances = initialBalances[address];
              let assetHashes = Object.keys(balances).map((assetID) => {
                return Util.sha3(assetID + balances[assetID]);
              });
              let assetTree = new IndexedMerkleTree(stageHeight, assetHashes);
              let roothash = Util.sha3(address + assetTree.rootHash);
              acc[address] = roothash;
              return acc;
            }, {});
          let accountTree = new IndexedMerkleTree(stageHeight, Object.values(accountHash));
          return { ok: true, receiptTree: receiptTree, accountTree: accountTree };
        }
      }
    };
    auditor = new Auditor({}, infinitechain);
  });

  afterAll(() => {
  });

  describe('#audit', () => {
    it('all checks should return empty array', async () => {
      let auditResult = await auditor.audit(stageHeight);
      expect(auditResult.receiptsWithRepeatedGSN.length).toBe(0);
      expect(auditResult.receiptsWithWrongBalance.length).toBe(0);
      expect(auditResult.receiptsWithSkippedGSN.length).toBe(0);
      expect(auditResult.receiptWithoutIntegrity.length).toBe(0);
    });
    it('should return once mistake in receiptsWithRepeatedGSN', async () => {
      let auditResult = await auditor.audit(stageHeight, receiptsWithRepeatedGSN);
      expect(auditResult.receiptsWithRepeatedGSN.length).toBe(1);
    });
    it('should return once mistake in receiptsWithWrongBalance', async () => {
      let auditResult = await auditor.audit(stageHeight, receiptsWithWrongBalance);
      expect(auditResult.receiptsWithWrongBalance.length > 0).toBeTruthy();
    });
    it('should return once mistake in receiptsWithSkippedGSN', async () => {
      let auditResult = await auditor.audit(stageHeight, receiptsWithSkippedGSN);
      expect(auditResult.receiptsWithSkippedGSN.length).toBe(1);
    });
    it('should return once mistake in receiptWithoutIntegrity', async () => {
      let auditResult = await auditor.audit(stageHeight, receiptsWithoutIntegrity);
      expect(auditResult.receiptWithoutIntegrity.length).toBe(1);
    });
  });
  describe('#auditAddress', () => {
    it('all checks should return empty array', async () => {
      let auditResult = await auditor.auditAddress(address, stageHeight);
      expect(auditResult.receiptsWithRepeatedGSN.length).toBe(0);
      expect(auditResult.receiptsWithWrongBalance.length).toBe(0);
      expect(auditResult.receiptWithoutIntegrity.length).toBe(0);
    });
    it('should return once mistake in receiptsWithRepeatedGSN', async () => {
      let addressReceipts = receiptsWithRepeatedGSN.filter(receipt => {
        return receipt.lightTxData.from == address || receipt.lightTxData.to == address;
      });
      let auditResult = await auditor.auditAddress(address, stageHeight, addressReceipts);
      expect(auditResult.receiptsWithRepeatedGSN.length).toBe(1);
    });
    it('should return once mistake in receiptsWithWrongBalance', async () => {
      let addressReceipts = receiptsWithWrongBalance.filter(receipt => {
        return receipt.lightTxData.from == address || receipt.lightTxData.to == address;
      });
      let auditResult = await auditor.auditAddress(address, stageHeight, addressReceipts);
      expect(auditResult.receiptsWithWrongBalance.length > 0).toBeTruthy();
    });
    it('should return once mistake in receiptWithoutIntegrity', async () => {
      let addressReceipts = receiptsWithoutIntegrity.filter(receipt => {
        return receipt.lightTxData.from == address || receipt.lightTxData.to == address;
      });
      let auditResult = await auditor.auditAddress(address, stageHeight, addressReceipts);
      expect(auditResult.receiptWithoutIntegrity.length).toBe(1);
    });
  });
  describe('#_pushOrNew', () => {
    it('should push value to original object array', async () => {
      let obj = { a: [1] };
      obj = auditor._pushOrNew(obj, 'a', 2);
      expect(obj.a.length).toBe(2);
    });
    it('should create new object array include value', async () => {
      let obj = { a: [1] };
      obj = auditor._pushOrNew(obj, 'b', 3);
      expect(obj.b.length).toBe(1);
    });
  });
  describe('#_addOrNew', () => {
    it('should add value to original bignumber', async () => {
      let originalValue = new BN(5);
      let addValue = new BN(3);
      let obj = { value: originalValue };
      obj = auditor._addOrNew(obj, 'value', addValue);
      let result = originalValue.add(addValue);
      expect(obj.value.toString()).toBe(result.toString());
    });
    it('should create new bignumber as value', async () => {
      let obj = {};
      let value = new BN(3);
      obj = auditor._addOrNew(obj, 'value', value);
      expect(obj.value.toString()).toBe(value.toString());
    });
  });
  describe('#_sortReceipts', () => {
    it('should return current order receipts', async () => {
      let sortedReceipts = auditor._sortReceipts(Stage2);
      for (let i = 1; i < sortedReceipts.length; i++) {
        expect(parseInt(sortedReceipts[i].receiptData.GSN, 16) > parseInt(sortedReceipts[i-1].receiptData.GSN, 16)).toBe(true);
      }
    });
  });
  describe('_getReceiptsGroup', () => {
    it('should return current receipt group', async ()=> {
      let group = {};
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let receiptsGroup = auditor._getReceiptsGroup(sortedReceipts);
      for (let i = 0; i < sortedReceipts.length; i++) {
        if (sortedReceipts[i].lightTxData.from != 0)
          auditor._pushOrNew(group, sortedReceipts[i].lightTxData.from, sortedReceipts[i]);
        if (sortedReceipts[i].lightTxData.to != 0 && sortedReceipts[i].lightTxData.to != sortedReceipts[i].lightTxData.from)
          auditor._pushOrNew(group, sortedReceipts[i].lightTxData.to, sortedReceipts[i]);
      }
      let keys = Object.keys(group);
      keys.forEach(key => {
        expect(receiptsGroup[key].length).toBe(group[key].length);
      });
    });
    it('should return current receipt group by address', async ()=> {
      let group = {};
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let receiptsGroup = auditor._getReceiptsGroup(sortedReceipts, address);
      for (let i = 0; i < sortedReceipts.length; i++) {
        if (sortedReceipts[i].lightTxData.from != 0 && sortedReceipts[i].lightTxData.from == address)
          auditor._pushOrNew(group, sortedReceipts[i].lightTxData.from, sortedReceipts[i]);
        if (sortedReceipts[i].lightTxData.to != 0 && sortedReceipts[i].lightTxData.to != sortedReceipts[i].lightTxData.from && sortedReceipts[i].lightTxData.to == address)
          auditor._pushOrNew(group, sortedReceipts[i].lightTxData.to, sortedReceipts[i]);
      }
      let keys = Object.keys(group);
      keys.forEach(key => {
        expect(receiptsGroup[key].length).toBe(group[key].length);
      });
    });
  });
  describe('#_repeatedGSNFilter', () => {
    it('should return empty array', async () => {
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let res = await auditor._repeatedGSNFilter(sortedReceipts);
      expect(res.length).toBe(0);
    });
    it('should find repeat GSN error', async () => {
      let sortedReceipts = auditor._sortReceipts(receiptsWithRepeatedGSN);
      let res = await auditor._repeatedGSNFilter(sortedReceipts);
      expect(res.length).toBe(1);
    });
  });
  describe('#_skippedGSNFilter', () => {
    it('should return empty array', async () => {
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let res = await auditor._skippedGSNFilter(sortedReceipts);
      expect(res.length).toBe(0);
    });
    it('should find skip GSN error', async () => {
      let sortedReceipts = auditor._sortReceipts(receiptsWithSkippedGSN);
      let res = await auditor._skippedGSNFilter(sortedReceipts);
      expect(res.length).toBe(1);
    });
  });
  describe('#_wrongBalanceFilter', () => {
    it('should return empty array', async () => {
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let receiptsGroup = auditor._getReceiptsGroup(sortedReceipts);
      let res = await auditor._wrongBalanceFilter(receiptsGroup, initialBalances);
      expect(res.length).toBe(0);
    });
    it('should find wrong balance error', async () => {
      let sortedReceipts = auditor._sortReceipts(receiptsWithWrongBalance);
      let receiptsGroup = auditor._getReceiptsGroup(sortedReceipts);
      let res = await auditor._wrongBalanceFilter(receiptsGroup, initialBalances);
      expect(res.length > 0).toBeTruthy();
    });
  });
  describe('#_integrityFilter', () => {
    it('should return empty array', async () => {
      let sortedReceipts = auditor._sortReceipts(Stage2);
      let res = await auditor._integrityFilter(sortedReceipts, receiptTree);
      expect(res.length).toBe(0);
    });
    it('should find integrity error', async () => {
      let sortedReceipts = auditor._sortReceipts(receiptsWithoutIntegrity);
      let res = await auditor._integrityFilter(sortedReceipts, receiptTree);
      expect(res.length).toBe(1);
    });
  });
  describe('#_computeRootHashFromSlice', () => {
    it('should return current root hash', async () => {
      let receiptSlice = receiptTree.getSlice('2d0ed83aa2e9ef5d801c82947582b375fd4d76cc87d7de2a73378ab942e45595');
      let rootHash = await auditor._computeRootHashFromSlice(receiptSlice, stageHeight);
      expect('0x' + rootHash).toBe(infinitechain.contract.getStageRootHash()[0]);
    });
  });
});
