import InfinitechainBuilder from '../src/infinitechain-builder';
import LightTransaction from '../src/models/light-transaction';
import Receipt from '../src/models/receipt';
import Types from '../src/models/types';
import MockEnv from './env.mock';
import MockData from './mock-data';
import Util from '../src/utils/util';

describe('Client', () => {
  let ganache, infinitechain, contract, client;

  beforeAll(async () => {
    await MockEnv.startGringotts();
    ganache = await MockEnv.startGanache(7000);

    infinitechain = new InfinitechainBuilder().
      setNodeUrl(MockEnv.nodeUrl).
      setWeb3Url(ganache.web3Url).
      setSignerKey(MockEnv.key).
      setStorage('memory').
      build();

    await infinitechain.initialize();
    contract = infinitechain.contract;
    client = infinitechain.client;
  }, 20000);

  afterAll(() => {
    MockEnv.stopGringotts();
  });

  describe('#makeProposeDeposit', () => {
    it('should return correct lightTx format', async () => {
      contract.web3().eth.sendTransaction({
        from: MockEnv.gringottsAddress(), 
        to: contract.booster().options.address, 
        value: contract.web3().utils.toWei('1', 'ether'),
        gas: '4700000'
      });
      let lightTx = await client.makeProposeDeposit();
      expect( new LightTransaction(lightTx) ).toBeInstanceOf(LightTransaction);
    }, 10000);
  });
  describe('#sendLightTx', () => {
    it('should return correct receipt format', async () => {
      let lightTx = new LightTransaction(MockData.receipt);
      let receipt =  await client.sendLightTx(lightTx);
      expect( new Receipt(receipt) ).toBeInstanceOf(Receipt);
    });
  });
  describe('#saveReceipt, #getReceipt', () => {
    it('should save and get recceipt from DB successfully', async () => {
      let receipt = new Receipt(MockData.receipt);
      await client.saveReceipt(receipt);
      let receiptInDB = await client.getReceipt(receipt.receiptHash);
      expect(receiptInDB).toEqual(receipt.toJson());
    });
  });
  describe('#proposeTokenDeposit', () => {
    it('should return correct txHash', async (done) => {
      const TWX = await contract.erc20(ganache.TWXAddress);
      TWX.once('Approval', {
        filter: { _owner: '0x' + MockEnv.gringottsAddress() },
        toBlock: 'latest'
      }, async () => {
        const proposeData = {
          depositAddress: MockEnv.gringottsAddress(),
          depositValue: '10000',
          depositAssetAddress: ganache.TWXAddress
        };
        let txHash = await client.proposeTokenDeposit(proposeData);
        expect(await contract.isTxFinished(txHash)).toBeTruthy();
        done();
      });
      await TWX.methods.approve(ganache.BoosterAddress, contract.web3().utils.toWei('10000'))
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);
  });
  describe('#getProposeDeposit', () => {
    it('should return correct deposit lightTX', async () => {
      const depositLightTxs = await client.getProposeDeposit();
      let counter = 0;
      for (let i = 0; i < depositLightTxs.length; i++) {
        let response = await client.sendLightTx(depositLightTxs[i]);
        let depositReceiptJson = response;
        counter+=1;
        let depositReceipt = new Receipt(depositReceiptJson);
        await client.saveReceipt(depositReceipt);
      }
      expect(depositLightTxs).toHaveLength(counter);
    }, 30000);
  });
  describe('#makeLightTx', () => {
    it('should return correct lightTX format', async () => {
      const assetList = await client.fetchSupportedTokens();
      const assetAddress = assetList[0];
      const remittanceData = {
        from: MockEnv.gringottsAddress(),
        to: new Array(39).fill(0).join('') + '1',
        assetID: assetAddress,
        value: '0.001',
        fee: '0.001'
      };
      const metadata = 'Some data.';
      const lightTx = await client.makeLightTx(Types.remittance, remittanceData, metadata);
      expect( new LightTransaction(lightTx) ).toBeInstanceOf(LightTransaction);
    });
  });
  describe('#makeProposeWithdrawal, #fetchSupportedTokens', () => {
    it('should return correct lightTX format', async () => {
      const assetList = await client.fetchSupportedTokens();
      const assetAddress = assetList[0];
      let lightTx = await client.makeProposeWithdrawal({
        assetID: assetAddress,
        value: '1'
      });

      let receiptJson = {
        lightTxData: lightTx.lightTxData,
        lightTxHash: lightTx.lightTxHash,
        sig: lightTx.sig,
        receiptData: {
          stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
          GSN: '0000000000000000000000000000000000000000000000000000000000000015',
          fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
          toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
          lightTxHash: lightTx.lightTxHash,
          fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
          toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
        },
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let signedReceipt = MockEnv.gringottsSign(receipt);
      await contract.booster().methods.proposeWithdrawal(signedReceipt.toArray())
        .send({
          to: contract.booster().options.address, 
          from: MockEnv.gringottsAddress(), 
          gas: '4700000'
        });
      expect( new LightTransaction(lightTx) ).toBeInstanceOf(LightTransaction);
    }, 10000);
  });
  describe('#getProposeWithdrawal', () => {
    it('should return correct withdrawal logs ', async () => {
      let withdrawLogs = await client.getProposeWithdrawal({});
      expect(withdrawLogs.length).toBe(1); 
    });
  });
  describe('#getBooterBalance', () => {
    it('should get correct booster balance ', async () => {
      const clientAddress = '5b41d422b28801e3af070c9d392fee8c79da744d';
      const assetID = '2C382B91405Af403ee2211D4237eC551e5fAC767';
      let boosterBalance = await client.getBooterBalance(clientAddress, assetID);
      expect(boosterBalance.balance).toBe('26238000000000000000000');
    });
  });
  describe('#_computeRootHashFromSlice', () => {
    it('should get correct rootHash from slice ', async () => {
      const data = await infinitechain.gringotts.getSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031');
      const slice = data.slice.slice;
      let rootHash = await client._computeRootHashFromSlice(slice, 1);
      expect(rootHash).toBe('e0d0850f4a4b7c417ac4c366920003c00966d1c8af4fef73e2d7501918cd674b');
    });
  });
  describe('#getSlice', () => {
    it('should get correct slice', async () => {
      let clientSlice = await client.getSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031');
      expect(clientSlice.slice).toEqual(MockData.sliceByStageHeightAndRecepitHash);
    });
  });
  describe('#auidtReceiptSlice', () => {
    it('should get false if not attached', async () => {
      let clientSlice = await client.getSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031');
      let result = await client.auidtReceiptSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031', clientSlice.slice);
      expect(result).toBeFalsy();
    });
    it('should get true if attached', async () => {
      contract.getStageRootHash = function (){return ['0xe0d0850f4a4b7c417ac4c366920003c00966d1c8af4fef73e2d7501918cd674b', ''];};
      let clientSlice = await client.getSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031');
      let result = await client.auidtReceiptSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031', clientSlice.slice);
      expect(result).toBeTruthy();
    });
  });
  describe('#_prepare', () => {
    it('should modify fee and from to 0 if it\'s deposit', async () => {
      let lightTxData = await client._prepare('0', MockData.receipt.lightTxData);
      expect(lightTxData.from).toBe('0');
      expect(lightTxData.fee).toBe('0');
    });
    it('should modify to to 0 and logID to byte32 hex if it\'s withdrawal', async () => {
      let lightTxData = await client._prepare('1', MockData.receipt.lightTxData);
      expect(lightTxData.to).toBe('0');
      expect(Util.isHex(lightTxData.logID)).toBeTruthy();
      expect(Util.isByte32(lightTxData.logID)).toBeTruthy();
    });
    it('should modify logID to 0 if it\'s remittance', async () => {
      let lightTxData = await client._prepare('2', MockData.receipt.lightTxData);
      expect(lightTxData.logID).toBe('0');
    });
  });
  describe('#_getNonce', () => {
    it('should return byte32 hex string', async () => {
      let nonce = await client._getNonce();
      expect(Util.isHex(nonce)).toBeTruthy();
      expect(Util.isByte32(nonce)).toBeTruthy();
    });
  });
  describe('#signLightTx', () => {
    it('should return signed lightTx', async () => {
      let lightTxData = MockData.receipt.lightTxData;
      let lightTx = await client.makeLightTx(Types.deposit, lightTxData, MockData.receipt.metadata);
      lightTx.sig.clientLightTx = {};
      expect(lightTx.hasClientLightTxSig()).toBeFalsy();

      let signedLightTx = await client.signLightTx(lightTx);
      expect(signedLightTx.hasClientLightTxSig()).toBeTruthy();
    });
  });
  describe('#getAllReceiptHashes', () => {
    it('should get all receipt hash bt stage', async () => {
      let receiptHashes = await client.getAllReceiptHashes(1);
      expect(receiptHashes.length > 0).toBeTruthy();
    });
  });
});

// [Skip]
// takeObjection //抗議