import Contract from '../src/contract';
import MockEnv from './env.mock';

describe('Contract', () => {
  let ganache, contract, serializedTx, txHash;

  beforeAll(async () => {
    ganache = await MockEnv.startGanache(7001);

    contract = new Contract({ web3Url: ganache.web3Url }, {
      gringotts: { fetchBoosterAddress: () => { return { contractAddress: ganache.BoosterAddress, accountAddress: MockEnv.gringottsAddress() }; } },
      signer: { 
        getAddress: () => { return MockEnv.gringottsAddress(); },
        getPrivateKey: () => { return MockEnv.key; } }
    });
  }, 10000);

  describe('#fetchBoosterAddress', () => {
    it('should fetch booster address', async () => {
      await contract.fetchBoosterAddress();
      expect(contract._boosterContractAddress).toBe(ganache.BoosterAddress);
    });
    it('should fetch account address', async () => {
      expect(contract._boosterAccountAddress).toBe(MockEnv.gringottsAddress());
    });
  });
  describe('#fetchWebSocketConnection', () => {  
    it('should initialize web3 connection', async () => {
      expect(await contract.fetchWebSocketConnection()).toBeUndefined();
    });
  });
  describe('#boosterAccountAddress', () => {
    it('should initialize boosterAccountAddress', async () => {
      expect(contract.boosterAccountAddress()).toBeDefined();
    });
  });
  describe('#booster', () => {  
    it('should initialize contract interface', async () => {
      expect(contract.booster()).toBeDefined();
      expect(contract.booster().options.address.toLowerCase()).toBe(ganache.BoosterAddress);
    });
  });
  describe('#web3', () => {
    it('should initialize web3', async () => {
      expect(contract.web3()).toBeDefined();
    });
  });
  describe('#erc20', () => {
    it('should initialize erc20 token', async () => {
      expect(contract.erc20('0x68c34a54ec562b2b6efc8e61c54f9314b93b1a44')).toBeTruthy();
    });
    it('should throw error if erc20 token is invalid', async () => {
      expect(() => { contract.erc20('0xc34a54ec562b2b6efc8e61c54f9314b93b1a44'); }).toThrow('ERC20 token address is invalid');
    });
  });
  describe('#proposeTokenDeposit', () => {
    it('should return token deposit txHash', async (done) => {
      const TWX = await contract.erc20(ganache.TWXAddress);
      TWX.once('Approval', {
        filter: { _owner: '0x' + MockEnv.gringottsAddress() },
        toBlock: 'latest'
      }, async () => {
        const proposeData = {
          depositAddress: MockEnv.gringottsAddress(),
          depositValue: '10000',
          depositAssetAddress: ganache.TWXAddress
        };
        txHash = await contract.proposeTokenDeposit(proposeData);
        expect(await contract.isTxFinished(txHash)).toBeTruthy();
        done();
      });
      await TWX.methods.approve(ganache.BoosterAddress, contract.web3().utils.toWei('10000'))
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);
  });
  describe('#_signRawTransaction', () => {
    it('should return serializedTx', async () => {
      const accounts = await contract.web3().eth.getAccounts();
      serializedTx = await contract._signRawTransaction(null, MockEnv.gringottsAddress(), accounts[1], 1);
      expect(serializedTx).toBeTruthy();
    });
  });
  describe('#_sendRawTransaction', () => {
    it('should return txHash', async () => {
      txHash = await contract._sendRawTransaction(serializedTx);
      expect(txHash).toBeTruthy();
    });
  });
  describe('#isTxFinished', () => {
    it('should return true if txHash is valid', async () => {
      expect(await contract.isTxFinished(txHash)).toBeTruthy();
    });
  });
  describe('#getStageRootHash', () => {
    it('should get empty array when get stage root hash before attach', async () => {
      const rootHashes = await contract.getStageRootHash(1);
      expect(rootHashes.receiptRootHash).toBe('0x0000000000000000000000000000000000000000000000000000000000000000');
      expect(rootHashes.accountRootHash).toBe('0x0000000000000000000000000000000000000000000000000000000000000000');
      expect(rootHashes.attachTimestamp).toBe('0');
    });
    it('should get stage root hash', async () => {
      const data = [
        '0x2db60ba250e1828f7f6164014a50453a3d4a1ec9682df990fba4638f6e616681', //
        '0x64e604787cbf194841e7b68d7cd28786f6c9a0a3ab9f8b0a0e87cb4387ab0107',
        '0x0', //ipfs address slice
        '0x0', //ipfs address slice
        '0x0', //ipfs address slice
        '0x0' //ipfs address slice
      ];
      await contract.booster().methods.attach(data, [], [])
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
      const rootHashes = await contract.getStageRootHash(1);
      expect(rootHashes.receiptRootHash).toBe(data[0]);
      expect(rootHashes.accountRootHash).toBe(data[1]);
      expect(parseInt(rootHashes.attachTimestamp)).toBeGreaterThan(0);
    });
  });
  describe('#isStageFinalized', () => {
    it('should return true if stage is finalized', async () => {
      const isFinalized = await contract.isStageFinalized(1);
      expect(isFinalized).toBeFalsy();
    });
    it('should return true if stage is finalized', async () => {
      await contract.booster().methods.finalizeStage()
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
      const isFinalized = await contract.isStageFinalized(1);
      expect(isFinalized).toBeTruthy();
    });
  });
});

/**
 * #ToDo:
 * challenge 
 * challengedWrongBalance 
 * compensate 
 * defend 
 * getObjectionableLightTxHashes 
 * getObjection 
 **/