const MockData = require('./mock-data');
const EthUtils = require('ethereumjs-util');
const nock = require('nock');
const ganache = require('ganache-cli');
const Web3 = require('web3');
const contractAbi = require('../src/abi/Booster.json');
const TWXAbi = require('../src/abi/TWX.json');

export default class MockEnv {
  static key = '41b1a0649752af1b28b3dc29a1556eee781e4a4c3a1f7f53f90fa834de098c4d';
  static BoosterAddress = '0x59d43ec586be07a9699d4f104f8b9f77367e6d50';
  static TWXAddress = '0xf7c7a5330759e8b7dfc60b2f4868248026512cf4';
  static nodeUrl = 'http://localhost:3000';
  static assetList = ['0x0000000000000000000000000000000000000000', '0x5b41d422b28801e3af070c9d392fee8c79da744d'];

  static sign (caller, object) {
    const klass = 'receipt';
    let hashKey = klass + 'Hash';
    let h = object[hashKey];
    let prefix = Buffer.from('\x19Ethereum Signed Message:\n32');
    let message = EthUtils.sha3(Buffer.concat([prefix, Buffer.from(h, 'hex')]));
    let sig = EthUtils.ecsign(message, Buffer.from(this.key, 'hex'));
  
    let postfix = klass.charAt(0).toUpperCase() + klass.slice(1);
    let sigKey = caller + postfix;
    object.sig[sigKey] = {
      r: '0x' + sig.r.toString('hex'),
      s: '0x' + sig.s.toString('hex'),
      v: '0x' + sig.v.toString(16).padStart(64, '0')
    };
    return object; 
  }

  static async startGanache (wsPort) {
    const web3Url = `ws://localhost:${wsPort}`;
    const key = `0x${MockEnv.key}`;
    const server = ganache.server({
      accounts: [{
        secretKey: key,
        balance: 10 ** 24
      }, '0x1'],
      ws: true,
      port: wsPort
    });
    server.listen(wsPort, () => {
      console.log('chain running on Port:', wsPort);
    });
    let web3 = new Web3(web3Url);
    const acc = MockEnv.gringottsAddress();
    const TWX = await new web3.eth.Contract(TWXAbi.abi)
      .deploy({ data: TWXAbi.bytecode })
      .send({ from: acc, gas: '4700000' });
    const booster = await new web3.eth.Contract(contractAbi.abi)
      .deploy({
        data: contractAbi.bytecode, 
        arguments: [acc, [TWX.options.address], 0]
      })
      .send({ from: acc, gas: '4700000' });
    return {
      web3Url: web3Url,
      TWXAddress: TWX.options.address.toLowerCase(),
      BoosterAddress: booster.options.address.toLowerCase()
    };
  }

  static gringottsAddress () {
    return EthUtils.privateToAddress(Buffer.from(this.key, 'hex')).toString('hex');
  }

  static gringottsSign (receipt) {
    return this.sign('booster', receipt);
  }

  static startGringotts () {
    nock(this.nodeUrl)
      .persist()
      .defaultReplyHeaders({ 'access-control-allow-origin': '*' })
      .get('/booster/address')
      .reply(200, { contractAddress: this.BoosterAddress, accountAddress: this.gringottsAddress() })
      .get('/viable/stage/height')
      .reply(200, { height: 1 })
      .get('/assetlist')
      .reply(200, { assetList: this.assetList })
      .get('/accounts/1')
      .reply(200, MockData.accountsByStageHeight)
      .get('/balance/5b41d422b28801e3af070c9d392fee8c79da744d')
      .reply(200, MockData.balanceByAccount)
      .get('/balance/5b41d422b28801e3af070c9d392fee8c79da744d?assetID=2C382B91405Af403ee2211D4237eC551e5fAC767')
      .reply(200, MockData.balanceByAccountAndAssetID)
      .get('/roothash/1')
      .reply(200, { ok: true, receiptRootHash: MockData.receiptRootHash, accountRootHash: MockData.accountRootHash })
      .get('/receipts/1')
      .reply(200, MockData.receiptByStageHeight)
      .get('/receipts/1?address=9e5d51dab54e05a7e085d2936bde152560ece292')
      .reply(200, MockData.receiptByStageHeightAndAddress)
      .get('/receipt_by_gsn/1')
      .reply(200, MockData.receiptByGsn)
      .get('/trees/1')
      .reply(200, { ok: true, receiptTree: MockData.receiptTree, accountTree: MockData.accountTree })
      .get('/slice/1/a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031')
      .reply(200, { ok: true, slice: MockData.sliceByStageHeightAndRecepitHash })
      .post('/send/light_tx')
      .reply(200, { ok: true, receipt: MockData.receipt });
  }

  static stopGringotts () {
    nock.cleanAll();
  }
}
