import InfinitechainBuilder from '../src/infinitechain-builder';
import LightTransaction from '../src/models/light-transaction';
import Receipt from '../src/models/receipt';
import MockEnv from './env.mock';

describe('event', () => {
  let ganache, contract, event, infinitechain;

  beforeAll(async () => {
    await MockEnv.startGringotts();
    ganache = await MockEnv.startGanache(7002);

    infinitechain = new InfinitechainBuilder().
      setNodeUrl(MockEnv.nodeUrl).
      setWeb3Url(ganache.web3Url).
      setSignerKey(MockEnv.key).
      setStorage('memory').
      build();

    await infinitechain.initialize();
    contract = infinitechain.contract;
    event = infinitechain.event;
  }, 10000);

  afterAll(() => {
    MockEnv.stopGringotts();
  });

  describe('#proposeDeposit', () => {
    it('should receive ProposeDeposit event when deposit', async (done) => {
      const TWX = await contract.erc20(ganache.TWXAddress);
      TWX.once('Approval', {
        filter: { _owner: '0x' + MockEnv.gringottsAddress() },
        toBlock: 'latest'
      }, async () => {
        const proposeData = {
          depositAddress: MockEnv.gringottsAddress(),
          depositValue: '10000',
          depositAssetAddress: ganache.TWXAddress
        };
        event.once.onProposeDeposit((error, events) => {
          expect(error).toBeNull();
          expect(events.event).toEqual('ProposeDeposit');
          done();
        });
        await contract.proposeTokenDeposit(proposeData);
      });

      await TWX.methods.approve(ganache.BoosterAddress, contract.web3().utils.toWei('10000'))
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);

    it('should receive ProposeDeposit events', (done) => {
      event.getProposeDeposit((error, events) => {
        expect(error).toBeNull();
        expect(events.length).toBe(1);
        expect(events[0].event).toEqual('ProposeDeposit');
        done();
      });
    });
  });

  describe('#Deposit', () => {
    it('should receive Deposit event when deposit', async (done) => {

      let lightTxJson = {
        lightTxData: {
          from: '0',
          to: MockEnv.gringottsAddress(),
          value: '10000',
          fee: '0',
          nonce: '1',
          assetID: ganache.TWXAddress,
          logID: 1
        }, metadata: ''
      };

      let lightTx = new LightTransaction(lightTxJson);
      let signedLightTx = infinitechain.signer.signWithClientKey(lightTx);

      let receiptData = {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: signedLightTx.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      };

      let receiptJson = {
        lightTxData: signedLightTx.lightTxData,
        lightTxHash: signedLightTx.lightTxHash,
        sig: signedLightTx.sig,
        receiptData: receiptData,
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let signedReceipt = MockEnv.gringottsSign(receipt);

      event.once.onDeposit((error, events) => {
        expect(error).toBeNull();
        expect(events.event).toEqual('VerifyReceipt');
        expect(events.returnValues._type).toEqual('0');
        done();
      });

      await contract.booster().methods.deposit(
        signedReceipt.toArray()
      ).send({ from: MockEnv.gringottsAddress(), gas: '4700000' });

    }, 10000);
  });

  describe('#onProposeWithdraw', () => {
    it('should receive ProposeWithdraw event when ProposeWithdraw', async (done) => {

      let lightTxJson = {
        lightTxData: {
          from: MockEnv.gringottsAddress(),
          to: '0',
          value: '10000',
          fee: '0',
          nonce: '1',
          assetID: ganache.TWXAddress,
          logID: 1
        }, metadata: ''
      };

      let lightTx = new LightTransaction(lightTxJson);
      let signedLightTx = infinitechain.signer.signWithClientKey(lightTx);

      let receiptData = {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: signedLightTx.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      };

      let receiptJson = {
        lightTxData: signedLightTx.lightTxData,
        lightTxHash: signedLightTx.lightTxHash,
        sig: signedLightTx.sig,
        receiptData: receiptData,
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let signedReceipt = MockEnv.gringottsSign(receipt);
      event.once.onProposeWithdrawal((error, events) => {
        expect(error).toBeNull();
        expect(events.event).toEqual('VerifyReceipt');
        expect(events.returnValues._type).toEqual('1');
        done();
      });
      await contract.booster().methods.proposeWithdrawal(
        signedReceipt.toArray()
      ).send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);
  });

  describe('#Attach', () => {
    it('should get Attach event when attach', async (done) => {
      const data = [
        '0x2db60ba250e1828f7f6164014a50453a3d4a1ec9682df990fba4638f6e616681',
        '0x64e604787cbf194841e7b68d7cd28786f6c9a0a3ab9f8b0a0e87cb4387ab0107',
        '0x0', //ipfs address slice
        '0x0', //ipfs address slice
        '0x0', //ipfs address slice
        '0x0' //ipfs address slice
      ];

      event.once.onAttach((error, events) => {
        expect(error).toBeNull();
        expect(events.event).toEqual('Attach');
        done();
      });

      await contract.booster().methods.attach(data, [], [])
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);
  });

  describe('#Finalize', () => {
    it('should receive Finalize event when stage is finalized', async (done) => {
      event.once.onWithdraw((error, events) => {
        expect(error).toBeNull();
        expect(events.event).toEqual('Withdraw');
      });

      event.once.onFinalize((error, events) => {
        expect(error).toBeNull();
        expect(events.event).toEqual('Finalize');
        done();
      });

      await contract.booster().methods.finalizeStage()
        .send({ from: MockEnv.gringottsAddress(), gas: '4700000' });
    }, 10000);
  });
});