import Gringotts from '../src/gringotts';
import MockEnv from './env.mock';
import MockData from './mock-data';
import LightTransaction from '../src/models/light-transaction';

describe('Gringotts', () => {
  beforeAll(() => {
    MockEnv.startGringotts();
  });

  afterAll(() => {
    MockEnv.stopGringotts();
  });

  describe('#fetchBoosterAddress', () => {
    it('should get correct address', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const boosterAddress = await gringotts.fetchBoosterAddress();

      expect(boosterAddress.contractAddress).toBe(MockEnv.BoosterAddress);
    });
  });

  describe('#fetchAssetList', () => {
    it('should get asset list', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const assetList = await gringotts.getAssetList();
      expect(assetList).toEqual(MockEnv.assetList);
    });
  });


  describe('#fetchViableStageHeight', () => {
    it('should get viable stage height', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const height = await gringotts.getViableStageHeight();
      expect(height).toBe(1);
    });
  });

  describe('#fetchAccount', () => {
    it('should get account by stage height', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getAccountBalances(1);
      expect(data).toEqual(MockData.accountsByStageHeight);
    });
  });

  describe('#fetchBalanceByAddress', () => {
    it('should get balance by address', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getBoosterBalance('5b41d422b28801e3af070c9d392fee8c79da744d');
      expect(data).toEqual(MockData.balanceByAccount);
    });
    it('should get account by stage height', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getBoosterBalance('5b41d422b28801e3af070c9d392fee8c79da744d', '2C382B91405Af403ee2211D4237eC551e5fAC767');
      expect(data).toEqual(MockData.balanceByAccountAndAssetID);
    });
  });

  describe('#fetchRootHashes', () => {
    it('should get root hash by stage', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.fetchRootHashes(1);
      expect(data.receiptRootHash).toBe(MockData.receiptRootHash);
      expect(data.accountRootHash).toBe(MockData.accountRootHash);
    });
  });

  describe('#getOffchainReceipt', () => {
    it('should get off chain receipt by GSN', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getOffchainReceiptByGSN(1);
      expect(data).toEqual(MockData.receiptByGsn);
    });

    it('should get off chain receipts by stage height', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getOffchainReceipts(1);
      expect(data).toEqual(MockData.receiptByStageHeight);
    });

    it('should get off chain receipts by stage height and Address', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getOffchainReceipts(1, '9e5d51dab54e05a7e085d2936bde152560ece292');
      expect(data).toEqual(MockData.receiptByStageHeightAndAddress);
    });
  });

  describe('#getTreesByStageHeight', () => {
    it('should get tree by stage', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getTrees(1);
      expect(data.receiptTree).toEqual(MockData.receiptTree);
      expect(data.accountTree).toEqual(MockData.accountTree);
    });
  });

  describe('#getSlice', () => {
    it('should get slice by stage height and receipt hash', async ()=> {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      const data = await gringotts.getSlice(1, 'a35c77ecffffcfc7a1f91bcac114b9eceac9fd415af37bcac68568a33f744031');
      expect(data.slice).toEqual(MockData.sliceByStageHeightAndRecepitHash);
    });
  });

  describe('#send light tranction', () => {
    it('shuould send lightTx', async () => {
      const gringotts = new Gringotts({ nodeUrl: MockEnv.nodeUrl });
      let tx = new LightTransaction(MockData.signedLightTx);
      const receipt = await gringotts.sendLightTx(tx);
      expect(receipt.toJson()).toEqual(MockData.receipt);
    });
  });
});
