import LightTransaction from '../../src/models/light-transaction';
import Receipt from '../../src/models/receipt';
import types from '../../src/models/types';

describe('Receipt', () => {
  describe('#constructor', () => {
    let lightTxJson = {
      lightTxData: {
        fee: '3',
        to: '0x456',
        from: '0x123',
        value: '100',
        nonce: '0x123',
        assetID: 1,
        logID: 1,
        foo: 'bar'
      },
      sig: {
        clientLightTx:{
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      }
    };

    let correctLightTx = new LightTransaction(lightTxJson);
    let correctReceiptData = {
      stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
      GSN: '0000000000000000000000000000000000000000000000000000000000000015',
      fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
      toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
      lightTxHash: correctLightTx.lightTxHash,
      fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
      toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
    };

    it('should check if lightTx"s clientLightTx signature is empty or not.', () => {
      let wrongSig = {
        clientLightTx: {}
      };

      let wrongLightTxJson = {
        lightTxData: lightTxJson.lightTxData,
        sig: wrongSig
      };

      let wrongLightTx = new LightTransaction(wrongLightTxJson);

      let receiptJson = {
        lightTxData: wrongLightTx.lightTxData,
        lightTxHash: wrongLightTx.lightTxHash,
        sig: wrongLightTx.sig,
        receiptData: correctReceiptData,
        metadata: ''
      };

      expect(() => { new Receipt(receiptJson); }).toThrow('\'clientLightTx\' signature is empty.');
    });

    it('should check if all receiptJsonKeys are included', () => {
      let wrongReceiptData = {
        stageHeight: 1,
        GSN: 21,
        lightTxHash:'12345',
        toBalance: 500,
        hello: 'hello',
        metadata: ''
      };

      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        sig: correctLightTx.sig,
        receiptData: wrongReceiptData,
        metadata: ''
      };

      expect(() => { new Receipt(receiptJson); }).toThrow('Parameter \'receiptJson\' does not include key \'lightTxHash\'.');
    });

    it('should check if all receiptDataKeys are included', () => {
      let wrongReceiptData = {
        stageHeight: 1,
        GSN: 21,
        fromPreGSN: 11,
        toPreGSN: 11,
        lightTxHash:'12345',
        toBalance: 500,
        hello: 'hello',
        metadata: ''
      };

      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: correctLightTx.sig,
        receiptData: wrongReceiptData,
        metadata: ''
      };

      expect(() => { new Receipt(receiptJson); }).toThrow('Parameter \'receiptData\' does not include key \'fromBalance\'.');
    });

    it('should return correct receipt', () => {
      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: correctLightTx.sig,
        receiptData: correctReceiptData,
        metadata: ''
      };

      let receipt = new Receipt(receiptJson);
      let result = {
        receiptHash: '283f92e41209ca5a91276b07e102b40a7cd216e481e9147066fdb59820651a94',
        receiptData: correctReceiptData
      };

      expect(receipt.receiptData).toEqual(result.receiptData);
      expect(receipt.receiptHash).toEqual(result.receiptHash);
    });
  });
  describe('#type', () => {
    it('should return correct receipt type', () => {
      let lightTxJson1 = {
        lightTxData: {
          fee: '3',
          to: '0x456',
          from: '0x123',
          value: '100',
          nonce: '0x123',
          assetID: 1,
          logID: 1,
          foo: 'bar'
        },
        sig: {
          clientLightTx:{
            v: 28,
            r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
            s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
          }
        }
      };
      let lightTxJson2 = {
        lightTxData: {
          fee: '3',
          to: '0x456',
          from: '0',
          value: '100',
          nonce: '0x123',
          assetID: 1,
          logID: 1,
          foo: 'bar'
        },
        sig: {
          clientLightTx:{
            v: 28,
            r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
            s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
          }
        }
      };
      let lightTxJson3 = {
        lightTxData: {
          fee: '3',
          to: '0',
          from: '0x123',
          value: '100',
          nonce: '0x123',
          assetID: 1,
          logID: 1,
          foo: 'bar'
        },
        sig: {
          clientLightTx:{
            v: 28,
            r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
            s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
          }
        }
      };
  
      let lightTx1 = new LightTransaction(lightTxJson1);
      let lightTx2 = new LightTransaction(lightTxJson2);
      let lightTx3 = new LightTransaction(lightTxJson3);

      let receiptData1 = {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: lightTx1.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      };
      let receiptData2 = {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: lightTx2.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      };
      let receiptData3 = {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: lightTx3.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      };

      let receiptJson1 = {
        lightTxData: lightTx1.lightTxData,
        lightTxHash: lightTx1.lightTxHash,
        sig: lightTx1.sig,
        receiptData: receiptData1,
        metadata: ''
      };
      let receiptJson2 = {
        lightTxData: lightTx2.lightTxData,
        lightTxHash: lightTx2.lightTxHash,
        sig: lightTx2.sig,
        receiptData: receiptData2,
        metadata: ''
      };
      let receiptJson3 = {
        lightTxData: lightTx3.lightTxData,
        lightTxHash: lightTx3.lightTxHash,
        sig: lightTx3.sig,
        receiptData: receiptData3,
        metadata: ''
      };

      let receipt1 = new Receipt(receiptJson1);
      let receipt2 = new Receipt(receiptJson2);
      let receipt3 = new Receipt(receiptJson3);

      expect(receipt1.type()).toBe(types.remittance);
      expect(receipt2.type()).toBe(types.deposit);
      expect(receipt3.type()).toBe(types.withdrawal);
    });
  });
  describe('#hasBoosterReceiptSig', () => {
    let lightTxJson = {
      lightTxData: {
        fee: '3',
        to: '0x456',
        from: '0x123',
        value: '100',
        nonce: '0x123',
        assetID: 1,
        logID: 1,
        foo: 'bar'
      },
      sig: {
        clientLightTx:{
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      }
    };

    let correctLightTx = new LightTransaction(lightTxJson);
    let correctReceiptData = {
      stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
      GSN: '0000000000000000000000000000000000000000000000000000000000000015',
      fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
      toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
      lightTxHash: correctLightTx.lightTxHash,
      fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
      toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
    };
    it('should retrun true if booster receipet is signed', () => {
      let sig = {
        clientLightTx: correctLightTx.sig.clientLightTx,
        boosterReceipt: {
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      };
      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: sig,
        receiptData: correctReceiptData,
        metadata: ''
      };

      let receipt = new Receipt(receiptJson);
      expect(receipt.hasBoosterReceiptSig()).toBeTruthy();
    });
  });
  describe('#toJson', () => {
    let lightTxJson = {
      lightTxData: {
        fee: '3',
        to: '0x456',
        from: '0x123',
        value: '100',
        nonce: '0x123',
        assetID: 1,
        logID: 1,
        foo: 'bar'
      },
      sig: {
        clientLightTx:{
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      }
    };

    let correctLightTx = new LightTransaction(lightTxJson);
    let correctReceiptData = {
      stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
      GSN: '0000000000000000000000000000000000000000000000000000000000000015',
      fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
      toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
      lightTxHash: correctLightTx.lightTxHash,
      fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
      toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
    };
    it('should return json type receipt', () => {
      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: correctLightTx.sig,
        receiptData: correctReceiptData,
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let expected = { 
        lightTxHash: 'ad60f4fda49c6ed68c8cd587bc5875cb36bebeb5e6465c85e9294c40367ca764',
        lightTxData:
        { from: '0000000000000000000000000000000000000000000000000000000000000123',
          to: '0000000000000000000000000000000000000000000000000000000000000456',
          assetID: '0000000000000000000000000000000000000000000000000000000000000001',
          value: '0000000000000000000000000000000000000000000000056bc75e2d63100000',
          fee: '00000000000000000000000000000000000000000000000029a2241af62c0000',
          nonce: '0000000000000000000000000000000000000000000000000000000000000123',
          logID: '0000000000000000000000000000000000000000000000000000000000000001',
          metadataHash: 'c5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470' },
        receiptHash: '283f92e41209ca5a91276b07e102b40a7cd216e481e9147066fdb59820651a94',
        receiptData:
        { stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
          GSN: '0000000000000000000000000000000000000000000000000000000000000015',
          fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
          toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
          lightTxHash: 'ad60f4fda49c6ed68c8cd587bc5875cb36bebeb5e6465c85e9294c40367ca764',
          fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
          toBalance: '00000000000000000000000000000000000000000000000000000000000001f4' },
        sig:
        { clientLightTx:
            { v: 28,
              r: '0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
              s: '0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709' },
        boosterReceipt: {} 
        },
        metadata: '' };
      expect(receipt.toJson()).toEqual(expected);
    });
  });
  describe('#toArray', () => {
    let lightTxJson = {
      lightTxData: {
        fee: '3',
        to: '0x456',
        from: '0x123',
        value: '100',
        nonce: '0x123',
        assetID: 1,
        logID: 1,
        foo: 'bar'
      },
      sig: {
        clientLightTx:{
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      }
    };

    let correctLightTx = new LightTransaction(lightTxJson);
    let correctReceiptData = {
      stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
      GSN: '0000000000000000000000000000000000000000000000000000000000000015',
      fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
      toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
      lightTxHash: correctLightTx.lightTxHash,
      fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
      toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
    };
    it('should return array type receipt', () => {
      let sig = {
        clientLightTx: correctLightTx.sig.clientLightTx,
        boosterReceipt: {
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      };
      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: sig,
        receiptData: correctReceiptData,
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let expected = [ 
        '0xad60f4fda49c6ed68c8cd587bc5875cb36bebeb5e6465c85e9294c40367ca764',
        '0x0000000000000000000000000000000000000000000000000000000000000123',
        '0x0000000000000000000000000000000000000000000000000000000000000456',
        '0x0000000000000000000000000000000000000000000000000000000000000001',
        '0x0000000000000000000000000000000000000000000000056bc75e2d63100000',
        '0x00000000000000000000000000000000000000000000000029a2241af62c0000',
        '0x0000000000000000000000000000000000000000000000000000000000000123',
        '0x0000000000000000000000000000000000000000000000000000000000000001',
        '0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470',
        28,
        '0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
        '0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709',
        '0x0000000000000000000000000000000000000000000000000000000000000015',
        '0x0000000000000000000000000000000000000000000000000000000000000010',
        '0x0000000000000000000000000000000000000000000000000000000000000012',
        '0x0000000000000000000000000000000000000000000000000000000000000032',
        '0x00000000000000000000000000000000000000000000000000000000000001f4',
        28,
        '0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
        '0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709' 
      ];
      expect(receipt.toArray()).toEqual(expected);
    });
  });
  describe('#toString', () => {
    let lightTxJson = {
      lightTxData: {
        fee: '3',
        to: '0x456',
        from: '0x123',
        value: '100',
        nonce: '0x123',
        assetID: 1,
        logID: 1,
        foo: 'bar'
      },
      sig: {
        clientLightTx:{
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      }
    };

    let correctLightTx = new LightTransaction(lightTxJson);
    let correctReceiptData = {
      stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
      GSN: '0000000000000000000000000000000000000000000000000000000000000015',
      fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
      toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
      lightTxHash: correctLightTx.lightTxHash,
      fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
      toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
    };
    it('should return string type receipt', () => {
      let sig = {
        clientLightTx: correctLightTx.sig.clientLightTx,
        boosterReceipt: {
          v: 28,
          r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
          s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
        }
      };
      let receiptJson = {
        lightTxData: correctLightTx.lightTxData,
        lightTxHash: correctLightTx.lightTxHash,
        sig: sig,
        receiptData: correctReceiptData,
        metadata: ''
      };
      let receipt = new Receipt(receiptJson);
      let expected = JSON.stringify(receipt.toJson());
      expect(receipt.toString()).toEqual(expected);
    });
  });
});
