import LightTransaction from '../src/models/light-transaction';
import Signer from '../src/signer';

const pk = '41b1a0649752af1b28b3dc29a1556eee781e4a4c3a1f7f53f90fa834de098c4d';

describe('Signer', () => {
  let sig = {
    clientLightTx:{
      v: 28,
      r:'0x384f9cb16fe9333e44b4ea8bba8cb4cb7cf910252e32014397c73aff5f94480c',
      s:'0x55305fc94b234c21d0025a8bce1fc20dbc7a83b48a66abc3cfbfdbc0a28c5709'
    }
  };

  let lightTx = new LightTransaction({
    lightTxData: {
      from: '0x123',
      to: '0x456',
      value: '100',
      nonce: '0x123',
      assetID: 1,
      logID: 1,
      fee: '3'
    },
    sig
  });

  describe('#importPrivateKey', () => {
    const signer = new Signer();
    it('should throw error when get wrong charactor', () => {
      expect(() => { signer.importPrivateKey('wrongKeyType'); }).toThrow('Private key is invalid.');
    });

    it('should has key value', () => {
      signer.importPrivateKey(pk);
      expect(JSON.stringify(signer.key)).toBe(JSON.stringify(Buffer.from(pk, 'hex')));
    });
  });

  describe('#hasPrivateKey', () => {
    const signer = new Signer();
    it('should not has key default', () => {
      expect(signer.hasPrivateKey()).toBeFalsy();
    });

    it('should has private key after import key', () => {
      signer.importPrivateKey(pk);
      expect(signer.hasPrivateKey()).toBeTruthy();
    });
  });

  describe('#getPrivateKey', () => {
    const signer = new Signer();

    it('should throw error when get wrong charactor', () => {
      expect(() => { signer.importPrivateKey('wrongKeyType'); }).toThrow('Private key is invalid.');
    });

    it('should return private key if input key', () => {
      expect(signer.getPrivateKey(pk)).toBe(pk);
    });

    it('should throw error if not input key and without key', () => {
      expect(() => { signer.getPrivateKey(); }).toThrow('ECC private key does not exist. Please generate or import your keypair.');
    });

    it('should get import key if not input key', () => {
      signer.importPrivateKey(pk);
      expect(signer.getPrivateKey()).toBe(pk);
    });
  });

  describe('#getPublicKey', () => {
    const signer = new Signer();
    it('should get public key if input key', () => {
      expect(signer.getPublicKey(pk)).toBe('7abf9e6bc38763af165e10bf1d293766c61b64ee3032c7fbd3345873e4dd5119a5cfa54f0ddbe9ae86787da833ba04f5c7682bd032991a3fffe38ab2b01e9a93');
    });

    it('should get import key if not input key', () => {
      signer.importPrivateKey(pk);
      expect(signer.getPublicKey()).toBe('7abf9e6bc38763af165e10bf1d293766c61b64ee3032c7fbd3345873e4dd5119a5cfa54f0ddbe9ae86787da833ba04f5c7682bd032991a3fffe38ab2b01e9a93');
    });
  });

  describe('#getAddress', () => {
    const signer = new Signer();
    it('should get address if input key', () => {
      expect(signer.getAddress(pk)).toBe('fb44fa0865747558066266061786e69336b5f3a2');
    });

    it('should get import key if not input key', () => {
      signer.importPrivateKey(pk);
      expect(signer.getAddress()).toBe('fb44fa0865747558066266061786e69336b5f3a2');
    });
  });

  describe('#signWithClientKey', () => {
    const signer = new Signer();
    it('checks object name', () => {
      let object = 'bar';
      expect(() => { signer.signWithClientKey(object, pk); }).toThrow('\'object\' should be instance of \'LightTransaction\'.');
    });

    it('checks private key', () => {
      let object = 'bar';
      expect(() => { signer.signWithClientKey(object); }).toThrow('ECC private key does not exist. Please generate or import your keypair.');
    });

    it('returns correct signature with private key', () => {
      let object = lightTx;
      let sig = signer.signWithClientKey(object, pk).sig;
      let result = {
        clientLightTx: {
          r: '0x18fa9cc76adda45a236a72c9dc7cfde3c673eefcd9cf70d811b8362889863043',
          s: '0x700446c703472c25e2623a48e1aeaeca35bfa57e3ca0188bd5baa429eb330804',
          v: '0x000000000000000000000000000000000000000000000000000000000000001b'
        }
      };
      expect(sig).toEqual(result);
    });

    it('returns correct signature without private key', () => {
      let object = lightTx;
      signer.importPrivateKey(pk);
      let sig = signer.signWithClientKey(object).sig;
      let result = {
        clientLightTx: {
          r: '0x18fa9cc76adda45a236a72c9dc7cfde3c673eefcd9cf70d811b8362889863043',
          s: '0x700446c703472c25e2623a48e1aeaeca35bfa57e3ca0188bd5baa429eb330804',
          v: '0x000000000000000000000000000000000000000000000000000000000000001b'
        }
      };
      expect(sig).toEqual(result);
    });
  });
});
