import IndexedMerkleTree from '../../src/utils/indexed-merkle-tree';

describe('IndexMerkleTree', () => {
  const stageHeight = 1;
  const leafElements = ['456', '123', '789', '000'];
  describe('#constructor', () => {
    let tree = new IndexedMerkleTree(stageHeight, leafElements);
    it('should return correct index merkle tree', () => {
      let expected = {
        stageHeight: 1,
        leafElements: [ '456', '123', '789', '000' ],
        treeHeight: 3,
        emptyNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
        treeNodes: [{
          treeNodeIndex: 4,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 5,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 6,
          treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62',
          treeNodeElements: ['456', '789'] },
        { treeNodeIndex: 7,
          treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9',
          treeNodeElements: ['000', '123'] },
        { treeNodeIndex: 2,
          treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' },
        { treeNodeIndex: 3,
          treeNodeHash: '6b97e7bc40ef71b07af4feb85da43d4dacb40a17312aa473d507397bb1dd6433' },
        { treeNodeIndex: 1,
          treeNodeHash: '1af81e437b88404e4468c690982cf14857890a3e4316807d89bc4c3dd2450e9c' }],
        rootHash: '4184b5e5d6c1b039558f9293e0a829deab76d469af0c7505e81b8ca36eed4c86' 
      };
      expect(tree).toEqual(expected);
    });
    describe('#_build', () => {
      it('should return correct treeNodes and rootHash', () => {
        let result = tree._build();
        let treeNodes = [{
          treeNodeIndex: 4,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 5,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 6,
          treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62',
          treeNodeElements: ['456', '789'] },
        { treeNodeIndex: 7,
          treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9',
          treeNodeElements: ['000', '123'] },
        { treeNodeIndex: 2,
          treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' },
        { treeNodeIndex: 3,
          treeNodeHash: '6b97e7bc40ef71b07af4feb85da43d4dacb40a17312aa473d507397bb1dd6433' },
        { treeNodeIndex: 1,
          treeNodeHash: '1af81e437b88404e4468c690982cf14857890a3e4316807d89bc4c3dd2450e9c' }];
        let rootHash = '4184b5e5d6c1b039558f9293e0a829deab76d469af0c7505e81b8ca36eed4c86';
        expect(result.treeNodes).toEqual(treeNodes);
        expect(result.rootHash).toEqual(rootHash);
      });   
    });
    describe('#_getLeafIndexRange', () => {
      it('should retuen correct index range', () => {
        let indexRange = tree._getLeafIndexRange();
        let expected = [ 4, 5, 6, 7 ];
        expect(indexRange).toEqual(expected);
      });
    });   
    describe('#_computeTreeHeight', () => {
      it('should return tree height = 1 if has 1 element', () => {
        let treeHeight = 1;
        expect(tree._computeTreeHeight(1)).toBe(treeHeight);
      });
      it('should return tree height = 2 if has 2~3 element', () => {
        let treeHeight = 2;
        expect(tree._computeTreeHeight(2)).toBe(treeHeight);
        expect(tree._computeTreeHeight(3)).toBe(treeHeight);
      });
      it('should return tree height = 3 if has 4~7 element', () => {
        let treeHeight = 3;
        expect(tree._computeTreeHeight(4)).toBe(treeHeight);
        expect(tree._computeTreeHeight(5)).toBe(treeHeight);
        expect(tree._computeTreeHeight(6)).toBe(treeHeight);
        expect(tree._computeTreeHeight(7)).toBe(treeHeight);
      });
    });
  });
  describe('#getSlice', () => {
    let tree = new IndexedMerkleTree(stageHeight, leafElements);
    it('should return correct slice by element(123)', () => {
      let expected = [{
        treeNodeIndex: 7,
        treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9' },
      { treeNodeIndex: 6,
        treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62' },
      { treeNodeIndex: 2,
        treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' 
      }];
      expect(tree.getSlice('123')).toEqual(expected);
    });
    it('should return correct slice by element(000)', () => {
      let expected = [{
        treeNodeIndex: 7,
        treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9' },
      { treeNodeIndex: 6,
        treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62' },
      { treeNodeIndex: 2,
        treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' 
      }];
      expect(tree.getSlice('000')).toEqual(expected);
    });
    it('should return correct slice by element(456)', () => {
      let expected = [{
        treeNodeIndex: 6,
        treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62' },
      { treeNodeIndex: 7,
        treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9' },
      { treeNodeIndex: 2,
        treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' 
      }];
      expect(tree.getSlice('456')).toEqual(expected);
    });
    it('should return correct slice by element(789)', () => {
      let expected = [{
        treeNodeIndex: 6,
        treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62' },
      { treeNodeIndex: 7,
        treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9' },
      { treeNodeIndex: 2,
        treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' 
      }];
      expect(tree.getSlice('789')).toEqual(expected);
    });
    describe('#_getSlice', () => {
      it('should return correct elements array by indexes [ 6, 2 ]', () => {
        let indexes = [ 6, 2 ];
        let expected = [{
          treeNodeIndex: 6,
          treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62',
          treeNodeElements: [ '456', '789' ] },
        { treeNodeIndex: 2,
          treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6'
        }];
        expect(tree._getSlice(indexes)).toEqual(expected);
      });
      it('should return correct elements array by indexes [ 1, 2, 3, 4, 5, 6, 7 ]', () => {
        let indexes = [ 1, 2, 3, 4, 5, 6, 7 ];
        let expected = [{
          treeNodeIndex: 4,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 5,
          treeNodeHash: 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          treeNodeElements: [] },
        { treeNodeIndex: 6,
          treeNodeHash: '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62',
          treeNodeElements: [ '456', '789' ] },
        { treeNodeIndex: 7,
          treeNodeHash: '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9',
          treeNodeElements: [ '000', '123' ] },
        { treeNodeIndex: 2,
          treeNodeHash: '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6' },
        { treeNodeIndex: 3,
          treeNodeHash: '6b97e7bc40ef71b07af4feb85da43d4dacb40a17312aa473d507397bb1dd6433' },
        { treeNodeIndex: 1,
          treeNodeHash: '1af81e437b88404e4468c690982cf14857890a3e4316807d89bc4c3dd2450e9c'
        }];
        expect(tree._getSlice(indexes)).toEqual(expected);
      });
    });
  });
  describe('#getAllLeafElements', () => {
    let tree = new IndexedMerkleTree(stageHeight, leafElements);
    it('should return correct elements array ["000", "123"]', () => {
      let expected = ['000', '123'];
      expect(tree.getAllLeafElements('000')).toEqual(expected);
      expect(tree.getAllLeafElements('123')).toEqual(expected);
    });
    it('should return correct elements array ["456", "789"]', () => {
      let expected = ['456', '789'];
      expect(tree.getAllLeafElements('456')).toEqual(expected);
      expect(tree.getAllLeafElements('789')).toEqual(expected);
    });
    describe('_getTreeNode', () => {
      it('should return correct treeNode by treeNodeIndex 7', () => {
        let treeNodeIndex = 7;
        let treeNode = {
          'treeNodeElements': ['000', '123'],
          'treeNodeHash': '59b6549436dd51183e87514bea5c254d2e39fbf24854dce235743f7857d3e8f9',
          'treeNodeIndex': 7
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 6', () => {
        let treeNodeIndex = 6;
        let treeNode = {
          'treeNodeElements': ['456', '789'],
          'treeNodeHash': '208a7c0c22c7eace94610d419ee97c5447ba7ecf697bcaff56096e4b9d64ff62',
          'treeNodeIndex': 6
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 5', () => {
        let treeNodeIndex = 5;
        let treeNode = {
          'treeNodeElements': [],
          'treeNodeHash': 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          'treeNodeIndex': 5
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 4', () => {
        let treeNodeIndex = 4;
        let treeNode = {
          'treeNodeElements': [],
          'treeNodeHash': 'f4915122cc2b3d2dae9e7d77a78c2eb0cef7b55de5c725eea56e2027332489e1',
          'treeNodeIndex': 4
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 3', () => {
        let treeNodeIndex = 3;
        let treeNode = {
          'treeNodeHash': '6b97e7bc40ef71b07af4feb85da43d4dacb40a17312aa473d507397bb1dd6433',
          'treeNodeIndex': 3
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 2', () => {
        let treeNodeIndex = 2;
        let treeNode = {
          'treeNodeHash': '7e61347ec07753f909ad97df2c4d6ee835159d07b41c565ac1ee03e5732618d6',
          'treeNodeIndex': 2
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
      it('should return correct treeNode by treeNodeIndex 1', () => {
        let treeNodeIndex = 1;
        let treeNode = {
          'treeNodeHash': '1af81e437b88404e4468c690982cf14857890a3e4316807d89bc4c3dd2450e9c',
          'treeNodeIndex': 1
        };
        expect(tree._getTreeNode(treeNodeIndex)).toEqual(treeNode);
      });
    });
  });
  describe('#computeLeafIndex', () => {
    let tree = new IndexedMerkleTree(stageHeight, leafElements);
    it('should return correct index in merkle tree with leafElement 000', () => {
      let expected = 7;
      expect(tree.computeLeafIndex('000')).toBe(expected);
    });
    it('should return correct index in merkle tree with leafElement 123', () => {
      let expected = 7;
      expect(tree.computeLeafIndex('123')).toBe(expected);
    });
    it('should return correct index in merkle tree with leafElement 456', () => {
      let expected = 6;
      expect(tree.computeLeafIndex('456')).toBe(expected);
    });
    it('should return correct index in merkle tree with leafElement 789', () => {
      let expected = 6;
      expect(tree.computeLeafIndex('789')).toBe(expected);
    });
  });
});