import LightTransaction from '../src/models/light-transaction';
import Receipt from '../src/models/receipt';
import Signer from '../src/signer';
import Verifier from '../src/verifier';

import MockEnv from './env.mock';

const signer = new Signer();
const verifier = new Verifier();

describe('Verifier', () => {
  let lightTxJson;
  let lightTx;
  let clientSignedLightTx;
  let receiptJson;
  let receipt;
  let boosterSignedReceipt;

  beforeAll(async () => {
    await MockEnv.startGringotts();
    signer.importPrivateKey(MockEnv.key);

    verifier._infinitechain = {
      contract: {
        boosterAccountAddress: () => { return MockEnv.gringottsAddress(); }
      }
    };

    lightTxJson = {
      lightTxData: {
        from: 'fb44fa0865747558066266061786e69336b5f3a2',
        to: 'fb44fa0865747558066266061786e69336b5f3a2',
        value: '0.5',
        fee: '0.1',
        nonce: '1',
        assetID: 1,
        logID: 1
      }
    };
    lightTx = new LightTransaction(lightTxJson);
    clientSignedLightTx = signer.signWithClientKey(lightTx);

    receiptJson = {
      lightTxData: lightTx.lightTxData,
      lightTxHash: lightTx.lightTxHash,
      sig: lightTx.sig,
      receiptData: {
        stageHeight: '0000000000000000000000000000000000000000000000000000000000000001',
        GSN: '0000000000000000000000000000000000000000000000000000000000000015',
        fromPreGSN: '0000000000000000000000000000000000000000000000000000000000000010',
        toPreGSN: '0000000000000000000000000000000000000000000000000000000000000012',
        lightTxHash: lightTx.lightTxHash,
        fromBalance: '0000000000000000000000000000000000000000000000000000000000000032',
        toBalance: '00000000000000000000000000000000000000000000000000000000000001f4'
      },
      metadata: ''
    };
    receipt = new Receipt(receiptJson);
    boosterSignedReceipt = MockEnv.gringottsSign(receipt);
  });

  afterAll(() => {
    MockEnv.stopGringotts();
  });

  describe('#_recover', () => {
    it('should return address when input hash and sign', () => {
      const address = signer.getAddress();
      expect(verifier._recover(clientSignedLightTx.lightTxHash, clientSignedLightTx.sig.clientLightTx)).toBe(address);
    });
  });

  describe('#_verify', () => {
    it('should throw error when does not get lTx or receipt', () => {
      expect(() => { verifier._verify({}); }).toThrow('\'object\' should be instance of \'LightTransaction\' or \'Receipt\'.');
    });

    it('should return correct lTx verification result', () => {
      expect(verifier._verify(clientSignedLightTx)).toBeTruthy();
    });

    it('should find lTx invalid if change lightTxHash', () => {
      let wrongLTx = new LightTransaction(lightTxJson);
      expect(verifier._verify(wrongLTx)).toBeTruthy();
      wrongLTx = signer.signWithClientKey(wrongLTx);
      wrongLTx.lightTxHash = '666';
      expect(verifier._verify(wrongLTx)).toBeFalsy();
    });

    it('should find lTx invalid if change from address', () => {
      let wrongLTx = new LightTransaction(lightTxJson);
      expect(verifier._verify(wrongLTx)).toBeTruthy();
      wrongLTx = signer.signWithClientKey(wrongLTx);
      wrongLTx.lightTxData.from = '0';
      expect(verifier._verify(wrongLTx)).toBeFalsy();
    });

    it('should find lTx invalid if change to address', () => {
      let wrongLTx = new LightTransaction(lightTxJson);
      expect(verifier._verify(wrongLTx)).toBeTruthy();
      wrongLTx = signer.signWithClientKey(wrongLTx);
      wrongLTx.lightTxData.to = '0';
      expect(verifier._verify(wrongLTx)).toBeFalsy();
    });

    it('should return correct receipt verification result', () => {
      expect(verifier._verify(boosterSignedReceipt)).toBeTruthy();
    });

    it('should find receipt invalid if has wrong lightData', () => {
      let wrongReceipt = new Receipt(receiptJson);
      wrongReceipt = MockEnv.gringottsSign(wrongReceipt);
      wrongReceipt.receiptHash = '555';
      expect(verifier._verify(wrongReceipt)).toBeFalsy();
    });
  });

  describe('#verifyLightTx', () => {
    it('should throw error get wrong input', () => {
      expect(() => { verifier.verifyLightTx({ foo: 'foo' }); }).toThrow('Parameter \'lightTx\' is not a LightTransaction instance.');
    });

    it('should returns correct verification result', () => {
      expect(verifier.verifyLightTx(clientSignedLightTx)).toBeTruthy();
    });
  });

  describe('#verifyReceipt', () => {
    it('should throw error get wrong input', () => {
      expect(() => { verifier.verifyReceipt({ foo: 'foo' }); }).toThrow('Parameter \'receipt\' is not a Receipt instance.');
    });

    it('should returns correct verification result', () => {
      expect(verifier.verifyReceipt(boosterSignedReceipt)).toBeTruthy();
    });
  });
});
